@@ -5,6 +5,10 @@
 '''
 This module contains a Pandas DataFrame wrapper and
 additional Pandas object helper functions.
+
+Version = one row of a dataframe, it has its _oid, _start and _end
+Object = specified by its _oid, one object can have multiple versions in the
+    result object.
 '''
 
 import logging
@@ -18,39 +22,13 @@ import pandas.tseries.offsets as off
 from pandas.tslib import Timestamp
 import pandas as pd
 import numpy as np
-from calendar import timegm
 
-from IPython.display import HTML
+from metriqueu.utils import dt2ts
 
 NUMPY_NUMERICAL = [np.float16, np.float32, np.float64, np.float128,
                    np.int8, np.int16, np.int32, np.int64]
 
 
-def mask_filter(f):
-    '''
-    Generic decorator for getting back filtered frame
-    data according to True/False mask filter frame matching
-
-    :param Pandas.DataFrame mask_frame:
-        DataFrame that maps index:True/False where True means it
-        matches the filter and False means it does not
-    :param bool filter_:
-        True will return back a DataFrame that contains only items
-        which matched the mask_frame filter. False returns back the
-        opposite.
-    '''
-    return decorator(_mask_filter, f)
-
-
-def _mask_filter(f, self, *args, **kwargs):
-    filter_ = args[-1]  # by convention, filter_ expected as last arg
-    mask_frame = f(self, *args, **kwargs)
-    if filter_ is None:
-        return mask_frame
-    else:
-        return type(self)(self[mask_frame == filter_])
-
-
 def filtered(f):
     '''
     Decorator function that wraps functions returning pandas
@@ -67,22 +45,22 @@ def filtered(f):
     return decorator(_filter, f)
 
 
-def to_timestamp(d):
-    ''' Convert a given datetime object to a utc timestamp value'''
-    return timegm(d.utctimetuple())
-
-
 class Result(DataFrame):
     ''' Custom DataFrame implementation for Metrique '''
-    def __init__(self, data=None):
+    def __init__(self, data=None, date=None):
         super(Result, self).__init__(data)
-        self._result_data = data
         # The converts are here so that None is converted to NaT
         self.to_datetime('_start')
         self.to_datetime('_end')
         self._lbound = self._rbound = None
+        self.set_date_bounds(date)
 
     def to_datetime(self, column):
+        '''
+        The json serialization/deserialization process leaves dates as
+        timestamps (in s).
+        This function converts the column to datetimes.
+        '''
         if column in self:
             if self[column].dtype in NUMPY_NUMERICAL:
                 self[column] = pd.to_datetime(self[column], unit='s')
@@ -135,64 +113,24 @@ class Result(DataFrame):
         else:
             return self.filter(before_end & after_start)
 
-    def _auto_select_scale(self, dts, start=None, end=None, ideal=300):
-        '''
-        Guess what a good timeseries scale might be,
-        given a particular data set, attempting to
-        make the total number of x values as close to
-        `ideal` as possible
-
-        This is a helper for plotting
-        '''
-        start = min(dts) if start is None else start
-        end = max(dts) if end is None else end
-        maximum_count = len(filter(lambda dt: start <= dt and dt <= end, dts))
-        daily_count = (end - start).days
-        if maximum_count <= ideal:
-            return 'maximum'
-        elif daily_count <= ideal:
-            return 'daily'
-        elif daily_count / 7 <= ideal:
-            return 'weekly'
-        elif daily_count / 30 <= ideal:
-            return 'monthly'
-        elif daily_count / 91 <= ideal:
-            return 'quarterly'
-        else:
-            return 'yearly'
-
-    def history(self, dates=None, counts=True,
-                predict_since=None, lin_reg_days=20):
+    def history(self, dates=None, linreg_since=None, lin_reg_days=20):
         '''
         Works only on a Result that has _start and _end columns.
-        most_recent=False should be set for this to work
 
         :param list dates: List of dates
-        :param bool counts: return counts (not ids)
-        :param datetime predict_since:
+        :param datetime linreg_since:
             estimate the future values using linear regression.
         :param integer lin_reg_days:
             Set how many past days should we use for prediction calulation
         '''
-        if dates is None:
-            dates = self.get_dates_range()
-
-        idx, vals = [], []
-        for dt in dates:
-            idx.append(dt)
-            if counts:
-                vals.append(self.on_date(dt, only_count=True))
-            else:
-                vals.append(list(self.on_date(dt)._oid))
-        ret = Series(vals, index=idx)
-        if predict_since is not None:
-            if not counts:
-                raise ValueError('counts must be True if predict_future_since'
-                                 'is not None.')
-            ret = self.predict_future(ret, predict_since, lin_reg_days)
+        dates = dates or self.get_dates_range()
+        vals = [self.on_date(dt, only_count=True) for dt in dates]
+        ret = Series(vals, index=dates)
+        if linreg_since is not None:
+            ret = self._linreg_future(ret, linreg_since, lin_reg_days)
         return ret.sort_index()
 
-    def predict_future(self, series, since, days=20):
+    def _linreg_future(self, series, since, days=20):
         '''
         Predicts future using linear regression.
 
@@ -209,17 +147,19 @@ class Result(DataFrame):
         last_days = pd.date_range(end=since, periods=days)
         hist = self.history(last_days)
 
-        xi = np.array([to_timestamp(d) for d in hist.index])
+        xi = np.array(map(dt2ts, hist.index))
         A = np.array([xi, np.ones(len(hist))])
         y = hist.values
         w = np.linalg.lstsq(A.T, y)[0]
 
         for d in series.index[series.index > since]:
-            series[d] = w[0] * to_timestamp(d) + w[1]
+            series[d] = w[0] * dt2ts(d) + w[1]
             series[d] = 0 if series[d] < 0 else series[d]
 
         return series
 
+    ############################# DATES RANGE ################################
+
     def get_dates_range(self, scale='auto', start=None, end=None):
         '''
         Returns a list of dates sampled according to the specified parameters.
@@ -238,24 +178,23 @@ class Result(DataFrame):
             Include start and end in the result if they are not included yet.
 
         '''
-        if scale not in ['auto', 'daily', 'weekly', 'monthly', 'quarterly',
-                         'yearly']:
+        if scale not in ['auto', 'maximum', 'daily', 'weekly', 'monthly',
+                         'quarterly', 'yearly']:
             raise ValueError('Incorrect scale: %s' % scale)
-        start = self._start.min() if start is None else Timestamp(start)
-        end = max(self._end.dropna().max(),
-                  self._start.max()) if end is None else Timestamp(end)
+        start = Timestamp(start or self._start.min())
+        end = Timestamp(end or max(Timestamp(self._end.max()),
+                                   self._start.max()))
         start = start if self.check_in_bounds(start) else self._lbound
         end = end if self.check_in_bounds(end) else self._rbound
 
-        if scale == 'auto' or scale == 'maximum':
+        if scale == 'auto':
+            scale = self._auto_select_scale(start, end)
+        if scale == 'maximum':
             start_dts = list(self._start.dropna().values)
             end_dts = list(self._end.dropna().values)
             dts = map(Timestamp, set(start_dts + end_dts))
             dts = filter(lambda ts: self.check_in_bounds(ts) and
                          ts >= start and ts <= end, dts)
-        if scale == 'auto':
-            scale = self._auto_select_scale(dts, start, end)
-        if scale == 'maximum':
             return dts
 
         freq = dict(daily='D', weekly='W', monthly='M', quarterly='3M',
@@ -264,33 +203,49 @@ class Result(DataFrame):
                       monthly=off.MonthEnd(), quarterly=off.QuarterEnd(),
                       yearly=off.YearEnd())
         ret = list(pd.date_range(start + offset[scale], end, freq=freq[scale]))
-        ret = [start] + ret if start < ret[0] else ret
-        ret = ret + [end] if end > ret[-1] else ret
+        ret = [start] + ret if ret and start < ret[0] else ret
+        ret = ret + [end] if ret and end > ret[-1] else ret
         ret = filter(lambda ts: self.check_in_bounds(ts), ret)
         return ret
 
-    ######################## FILTERS ##########################
-
-    def group_size(self, column, to_dict=False):
+    def _auto_select_scale(self, start=None, end=None, ideal=300):
         '''
-        Simply group items by the given column and return
-        dictionary (or Pandas Series) with each bucket size
-        '''
-        gby = self.groupby(column).size()
-        return gby.to_dict() if to_dict else gby
+        Guess what a good timeseries scale might be,
+        given a particular data set, attempting to
+        make the total number of x values as close to
+        `ideal` as possible
 
-    @mask_filter
-    def oids(self, oids, _filter=True):
-        ''' filter for only objects with matching object oids '''
-        # there *should* be an easier way to do this, without lambda...
-        # .. you could do oids.__contains__
-        return self['_oid'].map(lambda x: x in oids)
+        This is a helper for plotting
+        '''
+        start = start or self._start.min()
+        end = end or max(self._end.max(), self._start.max())
+        daily_count = (end - start).days
+        if daily_count <= ideal:
+            return 'daily'
+        elif daily_count / 7 <= ideal:
+            return 'weekly'
+        elif daily_count / 30 <= ideal:
+            return 'monthly'
+        elif daily_count / 91 <= ideal:
+            return 'quarterly'
+        else:
+            return 'yearly'
+
+    ############################### FILTERS ##################################
 
     @filtered
-    def unfinished(self):
+    def filter_oids(self, oids):
         '''
-        Leaves only those entities that has some version with _end equal None
-        or with _end larger than the right cutoff.
+        Leaves only objects with specified oids.
+        '''
+        oids = set(oids)
+        return self[self['_oid'].map(lambda x: x in oids)]
+
+    @filtered
+    def unfinished_objects(self):
+        '''
+        Leaves only versions of those objects that has some version with
+        `_end == None` or with `_end > right cutoff`.
         '''
         mask = self._end.isnull()
         if self._rbound is not None:
@@ -331,7 +286,7 @@ class Result(DataFrame):
             last[col_name] = age
             return last
 
-        cut_ts = datetime.utcnow() if self._rbound is None else self._rbound
+        cut_ts = self._rbound or datetime.utcnow()
         res = pd.concat([prep(df) for _, df in self.groupby(self._oid)])
         return res
 
@@ -385,7 +340,7 @@ class Result(DataFrame):
     @filtered
     def started_after(self, dt):
         '''
-        Leaves only those entities whose first version started after the
+        Leaves only those objects whose first version started after the
         specified date.
         '''
         dt = Timestamp(dt)
@@ -408,11 +363,17 @@ class Result(DataFrame):
         '''
         return pd.concat([function(df) for _, df in self.groupby(self._oid)])
 
-    ######################## Plotting #####################
+    ################################ SAVE ####################################
 
-    def plot_column(self, column, **kwargs):
-        x = self.group_size(column)
-        return x.plot(**kwargs)
+    def save_to_cube(self, oid, pyclient, cube='results', owner=None):
+        frame = self.to_dict('records')
+        obj = {'_oid': oid, 'frame': frame}
+        pyclient.cube_save([obj], cube=cube, owner=owner)
 
-    def as_html(self):
-        return HTML(self.to_html())
+
+#################################### LOAD ####################################
+
+def load_from_cube(oid, pyclient, cube='results', owner=None):
+    res = pyclient.find('_oid == %s' % oid,
+                        fields='__all__', cube=cube, owner=owner, raw=True)
+    return res
